import json
import os
import random
import socket
import sys
import time
import uuid
import torch
import numpy as np
import ignite

import pyklopp.metadata as pkmd

from tqdm import tqdm
from ignite.contrib.handlers import ProgressBar
from ignite.engine import create_supervised_trainer, create_supervised_evaluator, Events
from cleo import Command

from pyklopp import __version__, subpackage_import
from pyklopp.util import load_dataset_from_argument, save_paths_obtain_and_check, load_custom_config
from pyklopp.loading import load_modules, add_local_path_to_system


class TrainCommand(Command):
    """
    Trains a model

    train
        {model : Path to the pytorch model file}
        {dataset : Data set module for training}
        {--m|modules=* : Optional module file to load.}
        {--c|config=* : Configuration JSON string or file path.}
        {--s|save= : Path (including name) to save the model to}
    """

    def handle(self):
        # Early check for save path
        save_path_base, model_file_name = save_paths_obtain_and_check(self)

        # Model file path (a persisted .pth pytorch model)
        model_root_path = self.argument('model')
        if not os.path.exists(model_root_path):
            raise ValueError('Model not found in path "%s"' % model_root_path)

        # Add current absolute path to system path to load local modules
        # If initialized a module previously from a local module, then it must be available in path later again
        add_local_path_to_system(self.info)

        """
        Optional (local) module to load.
        There several functionalities can be bundled at one place.
        """
        modules_option = self.option('modules')
        loaded_modules = load_modules(modules_option)

        """
        Load dataset module file
        """
        dataset = None  # will be the dataset class, should be of type 'torch.utils.data.Dataset'
        fn_get_dataset = None  # optional function to load the dataset based on the allocated configuration
        class_dataset = None  # optional class which will be instanatiated with the configuration sub key 'dataset_config'

        # Either 'my_dataset' / 'my_dataset.py' or s.th. like 'torchvision.datasets.cifar.CIFAR10' or 'torchvision.datasets.mnist.MNIST'
        dataset_argument = str(self.argument('dataset'))

        """
        Assemble configuration with pre-defined config and user-defined json/file.
        """
        config = {
            'global_unique_id': str(uuid.uuid4()),
            'pyklopp_version': __version__,
            'loaded_modules': loaded_modules,
            'python_seed_initial': None,
            'python_seed_random_lower_bound': 0,
            'python_seed_random_upper_bound': 10000,
            'python_cwd': os.getcwd(),
            'hostname': socket.gethostname(),
            'time_config_start': time.time(),
            'model_root_path': model_root_path,
            'model_persistence_name': model_file_name,  # If later set to None/empty, model will not be persisted
            'save_path_base': save_path_base,
            'config_persistence_name': 'config.json',
            'config_key': 'training',
            'gpus_exclude': [],
            'gpu_choice': None,  # if None, then random uniform of all available is chosen
            'num_epochs': 10,
            'batch_size': 100,
            'learning_rate': 0.01,
            'argument_dataset': dataset_argument,
            'get_dataset_transformation': 'pyklopp.defaults.get_transform',
            'get_optimizer': 'pyklopp.defaults.get_optimizer',
            'get_loss': 'pyklopp.defaults.get_loss',
            'get_dataset_test': None,
        }
        # TODO metadata will replace config in future releases
        metadata = pkmd.init_metadata()
        metadata.system_loaded_modules = loaded_modules

        # Load user-defined configuration
        if self.option('config'):
            for config_option in self.option('config'):
                config_option = str(config_option)
                user_config = load_custom_config(config_option)
                config.update(user_config)

        # Dynamic configuration computations.
        # E.g. random numbers or seeds etc.

        # If desired set an initial (global) random seed
        if config['python_seed_initial'] is not None:
            random.seed(int(config['python_seed_initial']))

        # Generate a (local) initial seed, which might depend on the initial global seed
        # This enables reproducibility
        a = config['python_seed_random_lower_bound']
        b = config['python_seed_random_upper_bound']
        python_seed_local = random.randint(a, b)
        config['python_seed_local'] = python_seed_local if 'python_seed_local' not in config else config['python_seed_local']
        random.seed(config['python_seed_local'])
        np.random.seed(config['python_seed_local'])
        torch.manual_seed(config['python_seed_local'])
        torch.cuda.manual_seed(config['python_seed_local'])

        # Re-Check file path for persistence before going into training
        if config['model_persistence_name'] is not None and len(config['model_persistence_name']) > 0:
            model_file_name = config['model_persistence_name']
            model_file_path = os.path.join(save_path_base, model_file_name)
            if os.path.exists(model_file_path):
                raise ValueError('Model file path "%s" already exists' % model_file_path)

        config['time_config_end'] = time.time()

        """
        Load configured dataset for training.
        """
        self.info('Loading dataset.')
        dataset = load_dataset_from_argument(dataset_argument, config)

        config['dataset'] = str(dataset.__class__.__name__)
        n_training_samples = len(dataset)
        train_sampler = torch.utils.data.SubsetRandomSampler(np.arange(n_training_samples, dtype=np.int64))
        config['time_dataset_loading_start'] = time.time()
        train_loader = torch.utils.data.DataLoader(
            dataset,
            batch_size=config['batch_size'],
            sampler=train_sampler,
            num_workers=2
        )
        dataiter = iter(train_loader)
        initial_features, initial_labels = dataiter.next()
        config['time_dataset_loading_end'] = time.time()

        # Determine device to use (e.g. cpu, gpu:0, gpu:1, ..)
        if torch.cuda.is_available():
            cuda_no = config['gpu_choice']
            if cuda_no is None:
                cuda_no = np.random.choice(np.setdiff1d(np.arange(torch.cuda.device_count()), config['gpus_exclude']))
            elif cuda_no in config['gpus_exclude']:
                raise ValueError('Your configured GPU device number is in the exclusion list of your configuration.')
            device = torch.device('cuda:%s' % cuda_no)
        else:
            device = torch.device('cpu')
        config['device'] = str(device)

        # Load the model
        try:
            model = torch.load(model_root_path, map_location=device)
        except ModuleNotFoundError as e:
            raise ValueError('Could not find module when loading model: %s' % e)
        model.to(device)
        config['model_pythonic_type'] = str(type(model))

        fn_get_optimizer = subpackage_import(config['get_optimizer'])
        optimizer = fn_get_optimizer(model.parameters(), **config)
        config['optimizer'] = str(optimizer.__class__.__name__)

        fn_get_loss = subpackage_import(config['get_loss'])
        fn_loss = fn_get_loss(**config)
        config['loss'] = str(fn_loss.__class__.__name__)

        self.info('Configuration:')
        self.info(json.dumps(config, indent=2, sort_keys=True))

        """
        The training loop.
        """
        trainer = create_supervised_trainer(model, optimizer, fn_loss, device=device)

        pbar = ProgressBar(persist=True)
        pbar.attach(trainer, metric_names='all')

        metric_accuracy = ignite.metrics.Accuracy()
        metric_precision = ignite.metrics.Precision(average=False)
        metric_recall = ignite.metrics.Recall(average=False)
        metric_f1 = (metric_precision * metric_recall * 2 / (metric_precision + metric_recall)).mean()
        evaluation_metrics = {
            'accuracy': metric_accuracy,
            'precision': ignite.metrics.Precision(average=True),
            'recall': ignite.metrics.Recall(average=True),
            'f1': metric_f1,
            'loss': ignite.metrics.Loss(fn_loss)
        }
        evaluator = create_supervised_evaluator(
            model,
            metrics=evaluation_metrics,
            device=device
        )
        histories = {metric_name: [] for metric_name in evaluation_metrics}

        @trainer.on(Events.EPOCH_COMPLETED)
        def log_training_results(engine):
            evaluator.run(train_loader)
            metrics = evaluator.state.metrics

            metric_infos = []
            for metric_name in evaluation_metrics:
                histories[metric_name].append(metrics[metric_name])
                metric_infos.append("{metric}: {val:.3}".format(metric=metric_name, val=metrics[metric_name]))

            pbar.log_message(
                "Training - Epoch: {}, ".format(engine.state.epoch)
                + ', '.join(metric_infos)
            )

        config['time_model_training_start'] = time.time()
        trainer.run(train_loader, max_epochs=config['num_epochs'])
        config['time_model_training_end'] = time.time()

        # Add the collected histories after each completed epoch to config
        for metric_name in evaluation_metrics:
            config['training_%s' % metric_name] = histories[metric_name]

        """
        Evaluation
        """
        if 'get_dataset_test' in config and config['get_dataset_test'] is not None:
            fn_get_dataset_test = subpackage_import(config['get_dataset_test'])
            dataset_test = fn_get_dataset_test(**config)
            config['dataset_test'] = str(dataset_test.__class__.__name__)

            dataset_test_length = len(dataset_test)
            test_sampler = torch.utils.data.SubsetRandomSampler(np.arange(dataset_test_length, dtype=np.int64))
            test_loader = torch.utils.data.DataLoader(
                dataset_test,
                batch_size=config['batch_size'],
                sampler=test_sampler,
                num_workers=2
            )

            config['time_model_evaluation_start'] = time.time()
            evaluation_state = evaluator.run(test_loader)
            config['time_model_evaluation_end'] = time.time()

            for metric_name in evaluation_metrics:
                config['evaluation_%s' % metric_name] = np.float(evaluation_state.metrics[metric_name])

        """
        Optional configuration & model persistence.
        """
        if save_path_base is not None:
            if config['model_persistence_name'] is not None and len(config['model_persistence_name']) > 0:
                model_file_name = config['model_persistence_name']
                model_file_path = os.path.join(save_path_base, model_file_name)

                # If the model file path already exists, increment the name as long as we find a free name
                model_file_name_increment = 1
                model_file_name_parts = config['model_persistence_name'].split('.')
                model_file_name_ending = model_file_name_parts[-1]
                model_file_name_head = '.'.join(model_file_name_parts[:-1])
                while os.path.exists(model_file_path):
                    self.info('Model persistence path "%s" already exists. Choosing new one automatically.' % model_file_path)
                    model_file_name = '.'.join([model_file_name_head, model_file_name_increment, model_file_name_ending])
                    model_file_path = os.path.join(save_path_base, model_file_name)
                    model_file_name_increment += 1

                # Write model file name back in to config in case it changed during free-name-search
                config['model_persistence_name'] = model_file_name
                config['model_persistence_path'] = model_file_path

                self.info('Saving to "%s"' % model_file_path)
                config['time_model_save_start'] = time.time()
                torch.save(model, model_file_path)
                config['time_model_save_end'] = time.time()

            config_file_name = config['config_persistence_name']
            config_file_path = os.path.join(save_path_base, config_file_name)

            full_config = {}
            if os.path.exists(config_file_path):
                # Load possible existing config
                with open(config_file_path, 'r') as handle:
                    full_config = json.load(handle)

            config_key = config['config_key']
            if config_key not in full_config:
                # Add config in the dict with a sub-key, e.g. { 'train': {..} }
                full_config[config_key] = config
            else:
                # Add config in a list of the dict, e.g. { 'train': [{previous_config}, {..}] }
                if type(full_config[config_key]) is not list:
                    # Make sure, the config-key is a list, e.g. "{ 'train': {xyz} }" -> "{ 'train': [{xyz}] }"
                    full_config[config_key] = [full_config[config_key]]
                full_config[config_key].append(config)

            self.info('Writing configuration to "%s"' % config_file_path)
            with open(config_file_path, 'w') as handle:
                json.dump(full_config, handle, indent=2, sort_keys=True)

        self.info('Final configuration:')
        self.info(json.dumps(config, indent=2, sort_keys=True))
        self.info('Done.')



import json
import os
import socket
import time
import uuid
import warnings

import cleo
import torch

from pyklopp import __version__, subpackage_import
from pyklopp.loading import add_local_path_to_system


def count_parameters(model: torch.nn.Module):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def load_into_property_object(prop_obj, kwargs, prefix):
    for name in kwargs:
        setattr(prop_obj, prefix+name, kwargs[name])


def load_dataset_from_argument(dataset_arg: str, assembled_config: dict) -> torch.utils.data.Dataset:
    """

    :param dataset_arg:
    :param assembled_config:
    :return:
    """
    fn_get_dataset = None  # optional function to load the dataset based on the allocated configuration
    class_dataset = None  # optional class which will be instanatiated with the configuration sub key 'dataset_config'

    add_local_path_to_system()

    # For bash-completion, also allow the module name to end with .py and then simply remove it
    dataset_module_file_name = None
    dataset_possible_module_file_name = dataset_arg + '.py' if not dataset_arg.endswith('.py') else dataset_arg
    if os.path.exists(dataset_possible_module_file_name):
        # We've got an argument for which a local file exists, e.g. 'my_module' or 'my_module.py'
        dataset_module_file_name = dataset_possible_module_file_name
        dataset_module_name = dataset_module_file_name.replace('.py', '')

        try:
            module = __import__('.' + dataset_module_name, fromlist=[''])
        except ModuleNotFoundError:
            raise ModuleNotFoundError('Could not import "%s"' % dataset_module_name)

        try:
            fn_get_dataset = module.get_dataset
        except AttributeError:
            raise ValueError('Could not find dataset() function in your module "%s". You probably need to define get_dataset(**kwargs)' % dataset_module_name)

    # Assume argument to be s.th. like 'torchvision.datasets.cifar.CIFAR10'
    if dataset_module_file_name is None:
        if not '.' in dataset_arg:
            raise ValueError('Dataset must be an attribute of a module. Expecting at least one dot.')
        try:
            class_dataset = subpackage_import(dataset_arg)
        except ModuleNotFoundError:
            raise ValueError('Could not import %s' % dataset_arg)
        except AttributeError as e:
            raise ValueError('You passed <{arg}> but we could not find the attribute within your package. It has thrown an attribute error'.format(arg=dataset_arg), e)

    if fn_get_dataset is None and class_dataset is None:
        raise ValueError('Neither a dataset class nor a get_dataset(**kwargs) is defined.')

    if fn_get_dataset:
        # The preferred way to load a dataset: via a getter function my_module.get_dataset(**config) -> torch.utils.data.Dataset
        dataset = fn_get_dataset(**assembled_config)

        assembled_config['dataset_getter'] = fn_get_dataset.__name__
    else:
        # Assemble a dataset-class-only configuration dict
        dataset_config = {}
        if 'dataset_config' in assembled_config:
            # The special key 'dataset_config' may contain several keys for the dataset
            dataset_config = assembled_config['dataset_config']
        # All keys prefixed with 'dataset_' are also added (without prefix) to the config:
        for key in assembled_config:
            if key.startswith('dataset_'):
                ds_config_key = key.replace('dataset_', '')
                dataset_config[ds_config_key] = assembled_config[key]

        # In case of class instanatiation, try to load the custom transformation function
        if 'get_dataset_transformation' in assembled_config:
            get_dataset_transformation = assembled_config['get_dataset_transformation']
            try:
                fn_get_custom_transformation = subpackage_import(get_dataset_transformation)
            except ModuleNotFoundError:
                raise ValueError('Could not import transformation %s' % get_dataset_transformation)
            except AttributeError as e:
                raise ValueError('Could not load transformation due to attribute error for %s: %s' % (get_dataset_transformation, e))

            dataset_config['transform'] = fn_get_custom_transformation()

        try:
            dataset = class_dataset(**dataset_config)
        except TypeError as e:
            raise ValueError('Could not initialize dataset class.'
                             'Probably arguments are missing.'
                             'Arguments prefixed with "dataset_" are passed directly to the dataset object.'
                             'Collected dataset config was < {config} >'.format(config=dataset_config), e)

        assembled_config['dataset_class'] = class_dataset.__name__

    return dataset


def save_paths_obtain_and_check(command: cleo.Command) -> (str, str):
    # Early check for save path
    save_path_base = None
    model_file_name = None
    if command.option('save'):
        save_path = str(command.option('save'))
        model_file_name = os.path.basename(save_path)
        # TODO check for model file name
        save_path_base = os.path.dirname(save_path)
        if len(save_path_base) < 1:
            raise ValueError('You did not specify a valid save path. Given was "%s"' % save_path)
        if os.path.exists(os.path.join(save_path_base, model_file_name)):
            raise ValueError('Path "%s" already exists' % save_path_base)

        if not os.path.exists(save_path_base):
            os.makedirs(save_path_base)

    return save_path_base, model_file_name


def build_default_config(command: cleo.Command, base_config: dict = None):
    if base_config is None:
        base_config = {}
    config = base_config.copy()

    config.update({
        'global_unique_id': str(uuid.uuid4()),
        'pyklopp_version': __version__,
        'loaded_modules': None,
        'gpus_exclude': [],
        'gpu_choice': None,  # if None, then random uniform of all available is chosen
        'python_seed_initial': None,
        'python_seed_random_lower_bound': 0,
        'python_seed_random_upper_bound': 10000,
        'python_cwd': os.getcwd(),
        'hostname': socket.gethostname(),
        'time_config_start': time.time(),
        'save_path_base': None,
        'model_persistence_name': None,
        'config_persistence_name': 'config.json',
        'argument_dataset': None,
    })

    all_options = command.option()
    all_arguments = command.argument()
    for prefix_name, keywords in zip(['option', 'argument'], [all_options, all_arguments]):
        for kwname in keywords:
            config[prefix_name+'_'+kwname] = keywords[kwname]

    return config


def build_default_training_config(base_config: dict = None):
    if base_config is None:
        base_config = {}
    config = base_config.copy()

    config.update({
        'model_root_path': None,
        'num_epochs': 10,
        'batch_size': 100,
        'learning_rate': 0.01,
        'get_dataset_transformation': 'pyklopp.defaults.get_transform',
        'get_optimizer': 'pyklopp.defaults.get_optimizer',
        'get_loss': 'pyklopp.defaults.get_loss',
        'get_dataset_test': None,
    })

    return config


def load_custom_config(config_option: str):
    if os.path.exists(config_option):
        with open(config_option, 'r') as handle:
            user_config = json.load(handle)
    else:
        try:
            user_config = json.loads(config_option)
        except TypeError as e:
            raise ValueError('Invalid JSON as config passed.', e)
        except json.decoder.JSONDecodeError as e:
            raise ValueError('Your configuration can not decoded properly. It might contain invalid JSON: < {text} >'.format(text=config_option), e)

    if not type(user_config) is dict:
        raise ValueError('User config must be a dictionary, but <{type}> was given: < {text} >'.format(text=user_config, type=type(user_config).__name__))

    reserved_prefixes = ['time', 'option', 'argument', 'config', 'python']
    if any(map(lambda k: any((k.startswith(p) for p in reserved_prefixes)), user_config)):
        warnings.warn('Custom config contains a keyword beginning with a reserved prefix.'
                      'Reserved prefixes are {prefixes}'.format(prefixes=reserved_prefixes))

    return user_config





import os
import json
from typing import List

import time

import uuid

import jsonschema
import semantic_version as semver
from importlib_resources import files

from pyklopp import __version__

latest_version = '0.1.0'


def load_schema(version=None):
    if version is None:
        version = latest_version
    return json.loads(files('pyklopp').joinpath('schema/metadata-' + version + '.json').read_text())


def validate_schema(metadata: dict, schema=None, version=None):
    if schema is None:
        version = version if version is not None else metadata['schema_version'] if 'schema_version' in metadata else None
        schema = load_schema(version)

    try:
        jsonschema.validate(instance=metadata, schema=schema)
    except jsonschema.ValidationError as e:
        raise MetadataError('The given metadata does not match the expected schema.', e)

    return schema


class MetadataError(Exception):
   """Base class for other exceptions"""
   pass


class MappingError(MetadataError):
    def __init__(self, other: Exception, data: dict = None):
        super(MappingError, self).__init__(other)
        self._data = data

    def __str__(self):
        additional = ',\ndata=<%s>' % self._data if self._data is not None else ''
        return super(MappingError, self).__str__() + additional


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class MetadataReader(dict):
    _props = {}
    _path_json_file = None
    _schema = None
    _schema_version = None

    def __init__(self, version: str = None, path_json_file=None, schema: str = None, **kwargs):
        super().__init__(**kwargs)
        self._path_json_file = path_json_file
        self._schema_version = version
        self._schema = json.loads(schema) if schema is not None else None

    @property
    def path(self):
        return self._path_json_file

    @path.setter
    def path(self, path):
        if self._path_json_file is not None:
            raise ValueError('Use one metadata object per metadata file and do not overwrite objects with new paths.')
        self._path_json_file = path

    @property
    def schema_version(self):
        return self._schema_version

    def read(self, path=None, validate=True):
        if path is not None:
            self.path = path
        if self.path is None:
            raise ValueError('You need to specify the metadata path')

        with open(self.path, 'r') as read_handle:
            file_props = json.load(read_handle)

        if validate:
            self._schema = validate_schema(file_props)

        self._props = file_props

    def load(self, metadata: str, validate=True):
        props = json.loads(metadata)

        if validate:
            schema = self._schema
            schema_version = self._schema_version

            self._schema = validate_schema(props, schema=schema, version=schema_version)

        self._props = props

    def save(self, validate=True):
        if validate:
            validate_schema(self._props, schema=self._schema, version=self._schema_version)

        with open(self._path_json_file, 'w+') as write_handle:
            json.dump(self._props, write_handle)

    def __getitem__(self, item):
        if item in self._props:
            return self._props[item]
        raise MetadataError('Unknown metadata keyword <{name}>'.format(name=item))

    def __setitem__(self, key, value):
        self._props[key] = value

    def __getattr__(self, item):
        if item in self._props:
            value = self._props[item]
            return AttrDict(value) if type(value) is dict else value
        raise MetadataError('Unknown metadata keyword <{name}>'.format(name=item))


def with_annotated_properties(Cls):
    for attr_name in Cls.__annotations__:

        def getter(self, name=attr_name):
            class_prop_key = '_' + name
            return getattr(self, class_prop_key)

        def setter(self, value, name=attr_name):
            class_prop_key = '_' + name
            setattr(self, class_prop_key, value)

        prop = property(getter, setter)
        class_prop_key = '_' + attr_name
        #prop = property(lambda self: self._get_default(key, None)).setter(lambda self, value: self._set(key, value))
        setattr(Cls, attr_name, prop)
        setattr(Cls, class_prop_key, None)

    return Cls


class PropertyObject(object):
    def properties(self):
        return [name for name in dir(self) if not name.startswith('_') and not callable(getattr(self, name))]

    @classmethod
    def with_annotated_properties(clazz, Cls):
        for attr_name in Cls.__annotations__:

            def getter(self, name=attr_name):
                class_prop_key = '_' + name
                return getattr(self, class_prop_key)

            def setter(self, value, name=attr_name):
                class_prop_key = '_' + name
                setattr(self, class_prop_key, value)

            prop = property(getter, setter)
            class_prop_key = '_' + attr_name
            #prop = property(lambda self: self._get_default(key, None)).setter(lambda self, value: self._set(key, value))
            setattr(Cls, attr_name, prop)
            setattr(Cls, class_prop_key, None)

        return Cls

    @classmethod
    def build_fill_default(clazz):
        prop_obj = clazz()
        for attr_name in prop_obj.__annotations__:
            attr_type = prop_obj.__annotations__[attr_name]
            if attr_type == str:
                setattr(prop_obj, attr_name, '')
            elif attr_type == float:
                setattr(prop_obj, attr_name, 0.0)
            elif attr_type == int:
                setattr(prop_obj, attr_name, 0)
            elif attr_type == list:
                setattr(prop_obj, attr_name, [])
            else:
                name_default_getter = 'get_default'
                if hasattr(attr_type, name_default_getter):
                    type_def_value = getattr(attr_type, name_default_getter)
                    setattr(prop_obj, attr_name, type_def_value)
                else:
                    raise ValueError('Unknown type <{T}> and could not find a default getter <get> on it to specify a default value.'.format(T=attr_type, get=name_default_getter))
        return prop_obj


@PropertyObject.with_annotated_properties
class Metadata(PropertyObject):
    schema_version: str

    system_global_unique_id: str
    system_hostname: str
    system_pyklopp_version: str
    system_python_cwd: str
    system_python_seed_initial: int
    system_python_seed_local: int
    system_python_seed_random_lower_bound: int
    system_python_seed_random_upper_bound: int
    system_loaded_modules: list

    time_config_start: float
    time_config_end: float
    time_model_init_start: float
    time_model_init_end: float
    time_model_save_start: float
    time_model_save_end: float

    params_batch_size: int
    params_learning_rate: float
    params_device: str

    arguments_batch_size: int
    arguments_device: str
    arguments_dataset: str


class MetadataMap(object):
    def __init__(self, read_dict: dict, write_dict: dict):
        self._reader = read_dict
        self._writer = write_dict

    @property
    def specification(self) -> semver.SimpleSpec:
        raise NotImplementedError('Your mapping implementation needs to provide a requirement spec for which it agrees with schema versions.')

    def applies(self, version: semver.Version) -> bool:
        return self.specification.match(version)

    @property
    def reader(self):
        return self._reader

    @property
    def writer(self):
        return self._writer

    def __call__(self, data: PropertyObject):
        self.map_all(data)

    def map_all(self, data: PropertyObject, strict=False):
        for prop in data.properties():
            try:
                setattr(data, prop, self.map(prop, self.reader))
            except MappingError as e:
                if strict:
                    raise MappingError(e)

    def remap_all(self, data: PropertyObject):
        for prop in data.properties():
            self.remap(prop, getattr(data, prop), self.writer)

    def map(self, property_name: str, reader: dict) -> str:
        raise NotImplementedError

    def remap(self, property_name: str, value, write_to_dict: dict):
        raise NotImplementedError


class ScopedMetadataMap(MetadataMap):
    @property
    def scope_separator(self) -> str:
        return '_'

    def _get_scope_map(self):
        raise NotImplementedError('You need to define a mapper which translates a scoping from flat to dict')

    def map(self, metadata_key, reader: dict):
        scoped_parts = metadata_key.split(self.scope_separator)

        current_map = self._get_scope_map()
        prefix_idx = 0
        scopes = []
        unscoped_reader = reader
        while prefix_idx < len(scoped_parts):
            current_prefix = scoped_parts[prefix_idx]
            if current_prefix in current_map and current_prefix in unscoped_reader:
                unscoped_reader = unscoped_reader[current_prefix]
                current_map = current_map[current_prefix]
                scopes.append(current_prefix)
                prefix_idx += 1
            else:
                unscoped_key = self.scope_separator.join(scoped_parts[prefix_idx:])
                key = current_map[unscoped_key] if unscoped_key in current_map else unscoped_key
                if key not in unscoped_reader:
                    raise MappingError('Did not find <{name}> which was mapped to <{key}> with scope <{scope}>'.format(
                        name=metadata_key, key=key, scope=self.scope_separator.join(scopes)),
                        data=reader
                    )
                return unscoped_reader[key]

    def remap(self, metadata_key: str, value, write_to_dict: dict):
        scoped_parts = metadata_key.split(self.scope_separator)
        current_map = self._get_scope_map()
        unscoped = write_to_dict
        prefix_idx = 0
        while prefix_idx < len(scoped_parts):
            current_prefix = scoped_parts[prefix_idx]
            if current_prefix in current_map:
                current_map = current_map[current_prefix]
                if current_prefix not in unscoped:
                    unscoped[current_prefix] = {}
                unscoped = unscoped[current_prefix]
                prefix_idx += 1
            else:
                unscoped_key = self.scope_separator.join(scoped_parts[prefix_idx:])
                key = current_map[unscoped_key] if unscoped_key in current_map else unscoped_key
                unscoped[key] = value
                break


class MetadataMapV0V1(ScopedMetadataMap):
    _map_v0v1 = {
        'schema_version': 'schema_version',
        'system': 'system',
        'time': 'time',
        'arguments': 'arguments',
        'params': 'params'
    }

    @property
    def specification(self) -> semver.SimpleSpec:
        return semver.SimpleSpec('>=0.1.0,<0.2.0')

    def _get_scope_map(self):
        return self._map_v0v1


def init_metadata(**kwargs):
    m = Metadata.build_fill_default()
    m.schema_version = latest_version

    m.system_global_unique_id = str(uuid.uuid4())
    m.system_pyklopp_version = __version__
    m.system_python_cwd = os.getcwd()
    m.system_python_seed_initial = None
    m.system_python_seed_local = None
    m.system_python_seed_random_lower_bound = 0
    m.system_python_seed_random_upper_bound = 10000

    m.time_config_start = time.time()

    for name in kwargs:
        setattr(m, name, kwargs[name])

    return m


def get_mapping(version: semver.Version):
    raise NotImplementedError


def read_metadata(path_to_file: str):
    if not os.path.exists(path_to_file):
        raise FileNotFoundError('File not found: %s' % path_to_file)
    reader = MetadataReader(path_to_file)
    reader.read()
    version = reader.schema_version



import os
import sys


def subpackage_import(name: str):
    components = name.split('.')

    try:
        mod = __import__(components[0])
        for comp in components[1:]:
            mod = getattr(mod, comp)
    except AttributeError as e:
        raise AttributeError('Could not import <{name}>. System path info: "{syspath}"'.format(
            name=name, syspath=sys.path
        ), e)

    return mod


def add_local_path_to_system(fn_info=None):
    # Add current absolute path to system path to load local modules
    # If initialized a module previously from a local module, then it must be available in path later again
    add_path = os.path.abspath('.')
    if add_path not in sys.path:
        sys.path.append(add_path)

        if fn_info is not None:
            fn_info('Added "%s" to path.' % add_path)


def remove_local_path_from_system():
    remove_path = os.path.abspath('.')
    sys.path = [path for path in sys.path if path != remove_path]


def load_modules(module_args: list) -> list:
    loaded_modules = []

    if module_args is None:
        module_args = []

    for module_option in module_args:
        module_option = str(module_option)
        possible_module_file_name = module_option + '.py' if not module_option.endswith('.py') else module_option
        if os.path.exists(possible_module_file_name):
            module_file_name = possible_module_file_name
            module_name = module_file_name.replace('.py', '')

            try:
                loaded_modules.append(__import__('.' + module_name, fromlist=['']))
            except ModuleNotFoundError as e:
                raise ModuleNotFoundError('Could not import "%s". Have you added "." to your system path?' % module_name, e)

    return loaded_modules



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyklopp', 'pyklopp.console', 'pyklopp.console.commands']

package_data = \
{'': ['*'], 'pyklopp': ['schema/*']}

install_requires = \
['cleo>=0.7,<0.8',
 'importlib-metadata>=1.5.0,<2.0.0',
 'importlib-resources>=1.4.0,<2.0.0',
 'jsonschema>=3.2.0,<4.0.0',
 'pytorch-ignite>=0.2,<0.3',
 'semantic_version>=2.8.4,<3.0.0',
 'torch>=1.3',
 'torchvision>=0.4',
 'tqdm>=4.40,<5.0']

entry_points = \
{'console_scripts': ['pyklopp = pyklopp.console:main']}

setup_kwargs = {
    'name': 'pyklopp',
    'version': '0.3.0',
    'description': '',
    'long_description': '# pyklopp\nTired of logging all hyperparameter configurations of your model prototyping to disk?\n\nPyklopp is a tool to initialize, train and evaluate pytorch models (currently for supervised problems).\nIt persists all relevant hyperparameters, timings and model configurations.\nYour prototyping is reduced to defining your model, the dataset and your desired parameters.\n\n**Important note:** we are undergoing an architectural change from writing config json files to writing meta data files given a jsonschema.\nSo to keep your experiments reproducible and program against a current design of pyklopp, reference the exact pyklopp version in your experiment.\nE.g. for your *environment.yml*:\n```yaml\ndependencies:\n- pip:\n  - pyklopp==0.3.0\n```\n\n![Workflow sketch for developing a model and running it with pyklopp.](res/approach.png)\n\n\n## Installation\nYou can install a version from PyPi with: ``pip install pyklopp``.\n\nTo install the latest development build, you can clone the repository and invoke ``poetry build`` (having poetry installed).\nThen you can use the built package and install it with pip in your current environment by ``pip install dist/xxx.whl``.\n\n\n# Defining model & dataset\nSpecify your model in a plain python file, e.g.:\n```python\n# my_model.py\nimport torch.nn as nn\nimport torch.nn.functional as F\n\n\n# Your model can be any pytorch module\n# Make sure to not define it locally (e.g. within the get_model()-function)\nclass LeNet(nn.Module):\n    def __init__(self, output_size):\n        super(LeNet, self).__init__()\n        self.conv1 = nn.Conv2d(3, 6, 5)\n        self.conv2 = nn.Conv2d(6, 16, 5)\n        self.fc1 = nn.Linear(16 * 5 * 5, 120)\n        self.fc2 = nn.Linear(120, 84)\n        self.fc3 = nn.Linear(84, output_size)\n\n    def forward(self, x):\n        out = F.relu(self.conv1(x))\n        out = F.max_pool2d(out, 2)\n        out = F.relu(self.conv2(out))\n        out = F.max_pool2d(out, 2)\n        out = out.view(out.size(0), -1)\n        out = F.relu(self.fc1(out))\n        out = F.relu(self.fc2(out))\n        out = self.fc3(out)\n        return out\n\n\n# This is your model-instantiation function\n# It receives an assembled configuration keyword argument list and should return an instance of your model\ndef get_model(**kwargs):\n    output_size = int(kwargs[\'output_size\'])\n\n    return LeNet(output_size)\n```\n\nInvoke pyklopp to initialize it: ``pyklopp init my_model.get_model --save=\'test/model.pth\' --config=\'{"output_size": 10}\'``\nTrain it on *cifar10*:\n- ``pyklopp train test/model.pth cifar10.py --save=\'test/trained.pth\' --config=\'{"batch_size": 100}\'``\n- ``pyklopp train test/model.pth torchvision.datasets.cifar.CIFAR10 --save \'test/trained.pth\' --config=\'{"dataset_root": "/media/data/set/cifar10"}\'``\n\n\n# Examples\n\n```bash\n# Initializing & Saving: mymodel.py\npyklopp init foo --save=\'mymodel1/model.pth\'\npyklopp init foo --config=\'{"python_seed_initial": 100}\' --save=\'mymodel2/model.pth\'\n\n# Training\npyklopp train path/to/mymodel.pth mnist\npyklopp train path/to/mymodel.pth mnist --config=\'{"batch_size": 100, "learning_rate": 0.01}\'\n```\n\n```python\n# foo.py - Your model initialization function\nimport pypaddle.sparse\n\ndef init(**kwargs):\n    input_size = kwargs[\'input_size\']\n    output_size = kwargs[\'output_size\']\n\n    return pypaddle.sparse.MaskedDeepFFN(input_size, output_size, [100, 100])\n```\n\n```python\n# mnist.py - Your dataset loading functions\nimport pypaddle.util\n\n\ndef train_loader(**kwargs):\n    batch_size = kwargs[\'batch_size\']\n\n    mnist_train_loader, mnist_test_loader, _, selected_root = pypaddle.util.get_mnist_loaders(batch_size, \'/media/data/set/mnist\')\n    return mnist_train_loader\n\n\ndef test_loader(**kwargs):\n    batch_size = kwargs[\'batch_size\']\n\n    mnist_train_loader, mnist_test_loader, _, selected_root = pypaddle.util.get_mnist_loaders(batch_size, \'/media/data/set/mnist\')\n    return mnist_test_loader\n```\n\n\n# Development\n- Create wheel files in *dist/*: ``poetry build``\n- Install wheel in current environment with pip: ``pip install path/to/pyklopp/dist/pyklopp-0.1.0-py3-none-any.whl``\n\n## Running CI image locally\nInstall latest *gitlab-runner* (version 12.3 or up):\n```bash\n# For Debian/Ubuntu/Mint\ncurl -L https://packages.gitlab.com/install/repositories/runner/gitlab-runner/script.deb.sh | sudo bash\n\n# For RHEL/CentOS/Fedora\ncurl -L https://packages.gitlab.com/install/repositories/runner/gitlab-runner/script.rpm.sh | sudo bash\n\napt-get update\napt-get install gitlab-runner\n\n$ gitlab-runner -v\nVersion:      12.3.0\n```\nExecute job *tests*: ``gitlab-runner exec docker test-python3.6``',
    'author': 'Julian Stier',
    'author_email': 'julian.stier@uni-passau.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/innvariant/pyklopp',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import random
import socket
import sys
import os
import json
import time
import uuid
import torch.nn
import numpy as np
import pyklopp.metadata as pkmd

from cleo import Command
from pyklopp import __version__, subpackage_import
from pyklopp.util import count_parameters, save_paths_obtain_and_check, load_custom_config, load_into_property_object
from pyklopp.loading import load_modules, add_local_path_to_system


class InitCommand(Command):
    """
    Initializes a model from a given module

    init
        {model : Name of the module with initialization method for the model.}
        {--m|modules=* : Optional modules to load.}
        {--c|config=* : JSON config or path to JSON config file.}
        {--s|save= : Path to save the model to}
        {--meta= : Path to save the meta information to}
    """

    def handle(self):
        # Early check for save path
        save_path_base, model_file_name = save_paths_obtain_and_check(self)

        # Add current absolute path to system path
        # This is required for local modules to load.
        add_local_path_to_system(self.info)

        """
        Optional (local) module to load.
        There several functionalities can be bundled at one place.
        """
        modules_option = self.option('modules')
        loaded_modules = load_modules(modules_option)
        self.info('Loaded modules: %s' % loaded_modules)


        """ ---------------------------
        Static initial configuration.
        This can be overwritten by user-defined configurations.
        Dynamic computed configurations can be configured by some parameters and are computed after the user-defined
        values have been set.
        """
        config = {
            'global_unique_id': str(uuid.uuid4()),
            'pyklopp_version': __version__,
            'loaded_modules': loaded_modules,
            'gpus_exclude': [],
            'python_seed_initial': None,
            'python_seed_random_lower_bound': 0,
            'python_seed_random_upper_bound': 10000,
            'python_cwd': os.getcwd(),
            'hostname': socket.gethostname(),
            'time_config_start': time.time(),
            'save_path_base': save_path_base,
            'model_persistence_name': model_file_name,  # If later set to None/empty, model will not be persisted
        }
        # TODO metadata will replace config in future releases
        metadata = pkmd.init_metadata()
        metadata.system_loaded_modules = loaded_modules

        # Load user-defined configuration
        if self.option('config'):
            for config_option in self.option('config'):
                config_option = str(config_option)
                user_config = load_custom_config(config_option)
                config.update(user_config)
                load_into_property_object(metadata, user_config, 'arguments_')

        """ ---------------------------
        Dynamic configuration computations.
        E.g. random numbers or seeds etc.
        """

        # If desired set an initial (global) random seed
        if config['python_seed_initial'] is not None:
            random.seed(int(config['python_seed_initial']))

        # Generate a (local) initial seed, which might depend on the initial global seed
        # This enables reproducibility
        a = config['python_seed_random_lower_bound']
        b = config['python_seed_random_upper_bound']
        python_seed_local = random.randint(a, b)
        config['python_seed_local'] = python_seed_local if 'python_seed_local' not in config else config['python_seed_local']
        random.seed(config['python_seed_local'])
        np.random.seed(config['python_seed_local'])
        torch.manual_seed(config['python_seed_local'])
        torch.cuda.manual_seed(config['python_seed_local'])

        config['time_config_end'] = time.time()


        # For ease of usage, get_model can be either 'my_module.py' containing a get_model(**config) function
        # or it can be a function such as my_module.get_model with signature (**config) where my_module can be
        # a loadable module in the local path or an already loaded module via --modules
        get_model = self.argument('model')
        get_model_components = get_model.split('.')
        fn_get_model = None  # This will be the model loading function with signature fn(**config)

        if get_model.endswith('.py') or os.path.exists(get_model_components[0] + '.py'):
            module_name = get_model_components[0]
            module_file_name = module_name + '.py'
            if not os.path.exists(module_file_name):
                raise ValueError('No such local file %s' % module_file_name)

            try:
                module = __import__(module_name, fromlist=[''])
            except ModuleNotFoundError:
                raise ModuleNotFoundError('Could not import module "' + module_name + '"')

            if get_model.endswith('.py') or len(get_model_components) < 2:
                # get_model was 'my_module.py' or 'my_module'
                try:
                    fn_get_model = module.get_model
                except AttributeError:
                    raise ValueError(
                        'Could not find get_model() function in your module "%s". You probably need to define get_model(**config)' % module_name)
            else:
                # get_model was s.th. like 'my_module.load_model'
                mod = module
                for load_name in get_model_components[1:]:
                    mod = getattr(mod, load_name)
                fn_get_model = mod
        else:
            try:
                fn_get_model = subpackage_import(get_model)
            except ModuleNotFoundError as e:
                raise ValueError('Could not find module <%s> via subpackage import.'.format(get_model), e)

        config['argument_model'] = get_model
        config['get_model'] = str(fn_get_model.__name__)


        self.info('Configuration:')
        self.info(json.dumps(config, indent=2, sort_keys=True))

        # Instantiate model
        config['time_model_init_start'] = time.time()
        model = fn_get_model(**config)
        config['time_model_init_end'] = time.time()
        assert isinstance(model, torch.nn.Module)

        config['model_pythonic_type'] = str(type(model))
        config['model_trainable_parameters'] = count_parameters(model)

        if save_path_base is not None:
            if config['model_persistence_name'] is not None and len(config['model_persistence_name']) > 0:
                model_file_name = config['model_persistence_name']
                model_file_path = os.path.join(save_path_base, model_file_name)
                self.info('Saving to "%s"' % model_file_path)
                config['time_model_save_start'] = time.time()
                torch.save(model, model_file_path)
                config['time_model_save_end'] = time.time()

            config_file_name = 'config.json'
            config_file_path = os.path.join(save_path_base, config_file_name)
            self.info('Writing configuration to "%s"' % config_file_path)
            with open(config_file_path, 'w') as handle:
                json.dump({'init': config}, handle, indent=2, sort_keys=True)

        self.info('Final configuration:')
        self.info(json.dumps(config, indent=2, sort_keys=True))
        self.info('Done.')




import torchvision
import torch


def get_transform(**kwargs):
    return torchvision.transforms.Compose([
        torchvision.transforms.ToTensor()
    ])


def get_optimizer(parameters, learning_rate, **kwargs):
    return torch.optim.SGD(parameters, lr=learning_rate)


def get_loss(**kwargs):
    return torch.nn.CrossEntropyLoss()
